"""
Contains pytest configuration & fixtures.
"""
import pytest


@pytest.fixture
def portfolio_csv(tmp_path):
    """
    Creates a portfolio.csv in a temporary folder for the
    purposes of testing.
    """
    lines = [
        ('symbol,units,cost\r\n'),
        ('APPL,100,154.23\r\n'),
        ('AMZN,600,1223.43\r\n'),
    ]

    filename = tmp_path / 'portfolio.csv'
    with open(filename, 'w', newline='') as file:
        file.writelines(lines)

    return filename



"""
Generates performance reports for your stock portfolio.
"""
from pathlib import Path
import argparse
import csv
import sys
import requests

def main():
    """Entrypoint into programme"""
    args = get_args()
    #Only continues if arguments were given
    if sys.argv != 1:
        create_report(args)

def get_args(args=None):
    """Initialises and parses arguments from console window"""
    parser = argparse.ArgumentParser()
    parser.add_argument("-in", "--input", help="source of your portfolio")
    parser.add_argument("-out", "--output", help="destination where the report will be saved")
    args = parser.parse_args(args)
    #No arguments given will return the help menu
    if sys.argv == 1:
        parser.print_help()
    return args

def create_report(args):
    """Creates report from given CSV portfolio"""
    if args.input is None:
        print("An input file is required!\nPlease include an input CSV")
    else:
        try:
            input_path = string_to_path(args.input)
            report = read_portfolio(input_path)
            get_market_price(report)
            calculate_book_value(report)
            calculate_market_value(report)
            calculate_gain_loss(report)
            calculate_percent_change(report)
            #Saves report to user provided path or a default if none provided
            if args.output is None:
                save_report(report)
                print("Your report was saved at " + str(Path.cwd() / "new_report.csv"))
            else:
                output_path = string_to_path(args.output)
                save_report(report, filename=output_path)
                print("Your report was saved at " + str(output_path))
        except (requests.exceptions.HTTPError, requests.exceptions.RequestException):
            print("\nAn error has occurred in reaching the API.  Try again later")
        except FileNotFoundError:
            print("This is an invalid directory")

def string_to_path(filepath):
    """Creates a path variable from a user input string"""
    if not isinstance(filepath, str):
        raise ValueError
    #Path is rooted in current working directory
    path = Path.cwd() / filepath
    #Removes invalid filename characters
    name = "".join(char for char in path.name if char not in r'*|\/?<>:"')
    #Generates generic filename if empty
    if name == "":
        name = "new.csv"
    path = path.parent / name
    #Adds .csv suffix to filepaths if left out
    if path.suffix != ".csv":
        path = path.parent / (path.name + ".csv")
    return path

def dictionary_list_check(var):
    """Ensures that the list contains only dictionaries"""
    if not isinstance(var, list):
        raise ValueError
    removed = False
    #Temporarily stores items to be removed
    invalid = []
    for item in var:
        #Checks to see if list items are dictionaries removes others
        if not isinstance(item, dict):
            invalid.append(item)
            print(f"{item} was removed because of invalid formatting")
            removed = True
        #Checks to see if dictionary has requisite number of columns
        elif len(item) >= 3:
            #Checks all dictionary entries for missing data removes parent dictionary
            for entry in item:
                if item[entry] is None or item[entry] == "":
                    invalid.append(item)
                    print(f"{item} was removed because of missing value")
                    removed = True
        else:
            print("The input csv did not have enough columns")
            raise ValueError
    for item in invalid:
        var.remove(item)
    #Raises error if whole list had no dictionaries
    if len(var) == 0:
        print("The input list held no valid entries")
        raise ValueError
    #Prints warning if an item has been removed
    if removed:
        print(f"There are still {len(var)} item(s) in your report")

def read_portfolio(filename):
    """Returns data from a CSV file"""
    if not isinstance(filename, Path):
        raise ValueError
    with open(filename, newline='') as file:
        portfolio = []
        for holding in csv.DictReader(file):
            portfolio.append(holding)
        return portfolio

def get_market_price(portfolio):
    """Gets current price of holding from IEX"""
    dictionary_list_check(portfolio)
    #Creates list of symbols from user's portfolio
    symbols = [holding["symbol"] for holding in portfolio]
    #Attempts to reach the IEX API and get relevant stock prices
    try:
        uri = "https://api.iextrading.com/1.0/tops/last?symbols=" + ",".join(symbols)
        response = requests.get(uri)
        #Raises HTTP errors
        response.raise_for_status()
    except (requests.exceptions.HTTPError, requests.exceptions.RequestException) as error:
        print(error)
        raise error
    data = response.json()
    #Temporarily stores dictionaries with invalid symbols
    removed = []
    for holding in portfolio:
        #Attempts to find portfolio-data pair and marks for removal if none found
        try:
            price = next(entry for entry in data if entry["symbol"] == holding["symbol"])["price"]
            holding["latest_price"] = price
        except StopIteration:
            removed.append(holding)
            print(holding["symbol"] + " was not found; it is invalid!")
    for holding in removed:
        portfolio.remove(holding)

def calculate_book_value(report):
    """Calculates the initial value of the stocks"""
    dictionary_list_check(report)
    for holding in report:
        holding["book_value"] = round(int(holding["units"]) * float(holding["cost"]), 2)
    return report

def calculate_market_value(report):
    """Calculates the current value of the stocks"""
    dictionary_list_check(report)
    for holding in report:
        holding["market_value"] = round(int(holding["units"]) * float(holding["latest_price"]), 2)
    return report

def calculate_gain_loss(report):
    """Calculates the monetary value change"""
    dictionary_list_check(report)
    for holding in report:
        holding["gain_loss"] = round(
            float(holding["market_value"]) - float(holding["book_value"]), 2)
    return report

def calculate_percent_change(report):
    """Calculates the precentage value change"""
    dictionary_list_check(report)
    for holding in report:
        book_value = float(holding["book_value"])
        holding["change"] = round((float(holding["market_value"]) - book_value) / book_value, 2)
    return report

def save_portfolio(portfolio, filename=Path.cwd() / 'new_portfolio.csv'):
    """Saves data to a CSV file"""
    if not isinstance(filename, Path):
        raise ValueError
    dictionary_list_check(portfolio)
    with open(filename, "w", newline='') as file:
        #Creates then writes dictionary/csv header
        writer = csv.DictWriter(file, ['symbol', 'units', 'cost'])
        writer.writeheader()
        #Writes data to path
        writer.writerows(portfolio)

def save_report(report, filename=Path.cwd() / 'new_report.csv'):
    """Saves report to a CSV file"""
    if not isinstance(filename, Path):
        raise ValueError
    dictionary_list_check(report)
    with open(filename, "w", newline='') as file:
        #Creates then writes dictionary/csv header
        header = ['symbol', 'units', 'cost', 'latest_price',
                  'book_value', 'market_value', 'gain_loss', 'change']
        writer = csv.DictWriter(file, header)
        writer.writeheader()
        #Writes data to path
        writer.writerows(report)

if __name__ == '__main__':
    main()



# pylint: disable=missing-docstring
"""
The test module for portfolio_report
"""
from pathlib import Path
from collections import OrderedDict
import argparse
import pytest
import requests
from portfolio import portfolio_report

#read_portfolio test (functionality tested in test_io.py)
def test_read_non_path():
    with pytest.raises(ValueError):
        portfolio_report.read_portfolio("egg")

#save_portfolio test (functionality tested in test_io.py)
def test_save_port_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.save_portfolio("egg")

def test_save_port_non_path():
    with pytest.raises(ValueError):
        data = [{'symbol': 'MSFT', 'units': 10, 'cost': 99.66}]
        portfolio_report.save_portfolio(data, filename="egg")

#get_market_price_tests
def test_get_price_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.get_market_price("egg")

def test_get_price_404(requests_mock):
    with pytest.raises(requests.HTTPError):
        input_dict = [
            OrderedDict([
                ('symbol', 'ICE'),
                ('units', '600'),
                ('cost', '1223.43')
            ])
        ]
        requests_mock.get("https://api.iextrading.com/1.0/tops/last?symbols=ICE", status_code=404)
        portfolio_report.get_market_price(input_dict)

def test_get_price_500(requests_mock):
    with pytest.raises(requests.HTTPError):
        input_dict = [
            OrderedDict([
                ('symbol', 'ICE'),
                ('units', '600'),
                ('cost', '1223.43')
            ])
        ]
        requests_mock.get("https://api.iextrading.com/1.0/tops/last?symbols=ICE", status_code=500)
        portfolio_report.get_market_price(input_dict)

def test_get_price_connection(requests_mock):
    with pytest.raises(requests.ConnectionError):
        input_dict = [
            OrderedDict([
                ('symbol', 'ICE'),
                ('units', '600'),
                ('cost', '1223.43')
            ])
        ]
        uri = "https://api.iextrading.com/1.0/tops/last?symbols=ICE"
        requests_mock.get(uri, exc=requests.exceptions.ConnectionError)
        portfolio_report.get_market_price(input_dict)

def test_get_price_timeout(requests_mock):
    with pytest.raises(requests.Timeout):
        input_dict = [
            OrderedDict([
                ('symbol', 'ICE'),
                ('units', '600'),
                ('cost', '1223.43')
            ])
        ]
        uri = "https://api.iextrading.com/1.0/tops/last?symbols=ICE"
        requests_mock.get(uri, exc=requests.exceptions.Timeout)
        portfolio_report.get_market_price(input_dict)

def test_get_price_return(requests_mock):
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43')
        ])
    ]
    expected = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('price', '84.985'),
            ('size', '404'),
            ('time', '1565352854915')
        ])
    ]
    requests_mock.get("https://api.iextrading.com/1.0/tops/last?symbols=ICE", json=expected)
    portfolio_report.get_market_price(input_dict)
    assert float(input_dict[0]["latest_price"]) == 84.985

def test_get_price_invalid_symbol(requests_mock):
    input_dict = [
        OrderedDict([
            ('symbol', 'EGG'),
            ('units', '100'),
            ('cost', '154.23'),
        ]),
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43')
        ])
    ]
    expected = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('price', '84.985'),
            ('size', '404'),
            ('time', '1565352854915')
        ])
    ]
    requests_mock.get("https://api.iextrading.com/1.0/tops/last?symbols=EGG,ICE", json=expected)
    portfolio_report.get_market_price(input_dict)
    assert input_dict[0]["symbol"] == "ICE" and len(input_dict) == 1

#calculate_book_value tests
def test_book_value_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.calculate_book_value("egg")

def test_book_value_calculation():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43'),
            ('latest_price', '84.985')
        ])
    ]
    portfolio_report.calculate_book_value(input_dict)
    assert input_dict[0]["book_value"] == 734058

#calculate_market_value tests
def test_market_value_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.calculate_market_value("egg")

def test_market_value_calculation():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43'),
            ('latest_price', '84.985'),
            ('book_value', '734058.0')
        ])
    ]
    portfolio_report.calculate_market_value(input_dict)
    assert input_dict[0]["market_value"] == 50991

#calculate_gain_loss tests
def test_gain_loss_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.calculate_gain_loss("egg")

def test_gain_loss_calculation():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43'),
            ('latest_price', '84.985'),
            ('book_value', '734058.0'),
            ('market_value', '50991.0')
        ])
    ]
    portfolio_report.calculate_gain_loss(input_dict)
    assert input_dict[0]["gain_loss"] == -683067

#calculate_percent_change tests
def test_change_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.calculate_percent_change("egg")

def test_change_calculation():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43'),
            ('latest_price', '84.985'),
            ('book_value', '734058.0'),
            ('market_value', '50991.0'),
            ('gain_loss', '-683067.0')
        ])
    ]
    portfolio_report.calculate_percent_change(input_dict)
    assert input_dict[0]["change"] == -.93

#save_report tests
def test_save_rep_non_dict_list():
    with pytest.raises(ValueError):
        portfolio_report.save_report("egg")

def test_save_rep_non_path(portfolio_csv):
    with pytest.raises(ValueError):
        portfolio_report.save_report(portfolio_csv, filename="egg")

def test_save_report(portfolio_csv):
    data = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43'),
            ('latest_price', '84.985'),
            ('book_value', '734058.0'),
            ('market_value', '50991.0'),
            ('gain_loss', '-683067.0'),
            ('change', '-0.931')
        ])
    ]
    portfolio_report.save_report(data, filename=portfolio_csv)
    expected = "symbol,units,cost,latest_price,book_value,market_value,gain_loss,change\r\n"
    expected += "ICE,600,1223.43,84.985,734058.0,50991.0,-683067.0,-0.931\r\n"
    with open(portfolio_csv, "r", newline='') as file:
        result = file.read()
        assert result == expected, (
            f'Expecting the file to contain: \n{result}'
        )

#string_to_path tests
def test_non_string():
    with pytest.raises(ValueError):
        portfolio_report.string_to_path(3515380)

def test_string_to_path():
    path = portfolio_report.string_to_path("portfolio.csv")
    assert isinstance(path, Path) and path == Path.cwd() / "portfolio.csv"

def test_string_to_path_no_suffix():
    path = portfolio_report.string_to_path("portfolio")
    assert path == Path.cwd() / "portfolio.csv"

def test_string_to_path_bad_chars():
    path = portfolio_report.string_to_path(r'\/e:g*g?<l>e|g".csv')
    assert path.name == "eggleg.csv"

def test_string_to_path_all_bad():
    path = portfolio_report.string_to_path(r'*|\/?<>:"')
    assert path.name == "new.csv"

#dictionary_list_check tests
def test_dict_list_non_list():
    with pytest.raises(ValueError):
        portfolio_report.dictionary_list_check("egg")

def test_dict_list_all_bad():
    input_dict = [
        "egg",
        "leg"
    ]
    with pytest.raises(ValueError):
        portfolio_report.dictionary_list_check(input_dict)

def test_dict_list_some_bad():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43')
        ]),
        "egg",
        "leg"
    ]
    portfolio_report.dictionary_list_check(input_dict)
    assert len(input_dict) == 1 and input_dict[0]["symbol"] == "ICE"

def test_dict_list_all_good():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43')
        ]),
        OrderedDict([
            ('symbol', 'ENX'),
            ('units', '255'),
            ('cost', '404.01')
        ])
    ]
    portfolio_report.dictionary_list_check(input_dict)
    assert len(input_dict) == 2

def test_dict_list_missing_value():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43')
        ]),
        OrderedDict([
            ('symbol', 'ENX'),
            ('units', '255'),
            ('cost', None)
        ])
    ]
    portfolio_report.dictionary_list_check(input_dict)
    assert len(input_dict) == 1

def test_dict_list_empty_value():
    input_dict = [
        OrderedDict([
            ('symbol', 'ICE'),
            ('units', '600'),
            ('cost', '1223.43')
        ]),
        OrderedDict([
            ('symbol', 'ENX'),
            ('units', '255'),
            ('cost', '')
        ])
    ]
    portfolio_report.dictionary_list_check(input_dict)
    assert len(input_dict) == 1

def test_dict_list_missing_column():
    with pytest.raises(ValueError):
        input_dict = [
            OrderedDict([
                ('symbol', 'ICE'),
                ('units', '600')
            ]),
            OrderedDict([
                ('symbol', 'ENX'),
                ('units', '255')
            ])
        ]
        portfolio_report.dictionary_list_check(input_dict)

#get_args tests
def test_get_args_input():
    args = portfolio_report.get_args(["-in", "portfolio.csv"])
    assert args.input == "portfolio.csv"

def test_get_args_output():
    args = portfolio_report.get_args(["-out", "report.csv"])
    assert args.output == "report.csv"

#create_report tests
def test_create_report(portfolio_csv, requests_mock, tmp_path):
    returned = [
        OrderedDict([
            ('symbol', 'APPL'),
            ('price', '1478.19'),
            ('size', '404'),
            ('time', '1565352854915')
        ]),
        OrderedDict([
            ('symbol', 'AMZN'),
            ('price', '84.985'),
            ('size', '404'),
            ('time', '1565352854915')
        ])
    ]
    requests_mock.get("https://api.iextrading.com/1.0/tops/last?symbols=APPL,AMZN", json=returned)
    expected = "symbol,units,cost,latest_price,book_value,market_value,gain_loss,change\r\n"
    expected += "APPL,100,154.23,1478.19,15423.0,147819.0,132396.0,8.58\r\n"
    expected += "AMZN,600,1223.43,84.985,734058.0,50991.0,-683067.0,-0.93\r\n"
    portfolio_report.create_report(argparse.Namespace(input=str(portfolio_csv),
                                                      output=str(tmp_path / "report.csv")))
    with open((tmp_path / "report.csv"), "r", newline='') as file:
        result = file.read()
        assert result == expected, (
            f'Expecting the file to contain: \n{result}'
        )



"""
Tests I/O disk operations.
"""
from collections import OrderedDict

from portfolio import portfolio_report


# Note: the portfolio_csv argument found in the tests below
#       is a pytest "fixture". It is defined in conftest.py

# DO NOT edit the provided tests. Make them pass.

def test_read_portfolio(portfolio_csv):
    """
    Given that the read_portfolio is called, assert that
    the data the expected data is returned.
    """
    expected = [
        OrderedDict([
            ('symbol', 'APPL'),
            ('units', '100'),
            ('cost', '154.23'),
        ]),
        OrderedDict([
            ('symbol', 'AMZN'),
            ('units', '600'),
            ('cost', '1223.43')
        ])
    ]

    assert portfolio_report.read_portfolio(portfolio_csv) == expected, (
        'Expecting to get the data stored in the portfolio_csv '
        'fixture as a Python data structure.'
    )


def test_save_portfolio(portfolio_csv):
    """
    Given that the save portfolio method is called with the following
    data, assert that a CSV file is written in the expected format.

    The portfolio
    """
    data = [{'symbol': 'MSFT', 'units': 10, 'cost': 99.66}]
    portfolio_report.save_portfolio(data, filename=portfolio_csv)

    expected = 'symbol,units,cost\r\nMSFT,10,99.66\r\n'
    with open(portfolio_csv, 'r', newline='') as file:
        result = file.read()
        assert result == expected, (
            f'Expecting the file to contain: \n{result}'
        )



# pylint: disable=missing-docstring
import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="portfolio-report-Sealeon",
    version="1.0.3",
    author="Emily Knight",
    author_email="emily.knight2@mohawkcollege.ca",
    description="A stock report generator",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/sheridan-python/assignment-7-Sealeon",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "portfolio_report=portfolio.portfolio_report:main",
        ],
    },
)
